@@ -25,6 +25,7 @@
 #include "soup-connection.h"
 #include "soup-marshal.h"
 #include "soup-message.h"
+#include "soup-message-filter.h"
 #include "soup-misc.h"
 #include "soup-socket.h"
 #include "soup-ssl.h"
@@ -42,8 +43,11 @@ struct SoupConnectionPrivate {
 	SoupUri     *proxy_uri, *origin_uri, *conn_uri;
 	gpointer     ssl_creds;
 
+	SoupMessageFilter *filter;
+
 	SoupMessage *cur_req;
 	time_t       last_used;
+	gboolean     in_use;
 };
 
 #define PARENT_TYPE G_TYPE_OBJECT
@@ -65,6 +69,7 @@ enum {
   PROP_ORIGIN_URI,
   PROP_PROXY_URI,
   PROP_SSL_CREDS,
+  PROP_MESSAGE_FILTER,
 
   LAST_PROP
 };
@@ -95,6 +100,9 @@ finalize (GObject *object)
 	if (conn->priv->origin_uri)
 		soup_uri_free (conn->priv->origin_uri);
 
+	if (conn->priv->filter)
+		g_object_unref (conn->priv->filter);
+
 	g_free (conn->priv);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -193,6 +201,12 @@ class_init (GObjectClass *object_class)
 				      "SSL credentials",
 				      "Opaque SSL credentials for this connection",
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		object_class, PROP_MESSAGE_FILTER,
+		g_param_spec_pointer (SOUP_CONNECTION_MESSAGE_FILTER,
+				      "Message filter",
+				      "Message filter object for this connection",
+				      G_PARAM_READWRITE));
 }
 
 SOUP_MAKE_TYPE (soup_connection, SoupConnection, class_init, init, PARENT_TYPE)
@@ -254,6 +268,9 @@ set_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDS:
 		conn->priv->ssl_creds = g_value_get_pointer (value);
 		break;
+	case PROP_MESSAGE_FILTER:
+		conn->priv->filter = g_object_ref (g_value_get_pointer (value));
+		break;
 	default:
 		break;
 	}
@@ -278,11 +295,37 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDS:
 		g_value_set_pointer (value, conn->priv->ssl_creds);
 		break;
+	case PROP_MESSAGE_FILTER:
+		g_value_set_pointer (value, g_object_ref (conn->priv->filter));
+		break;
 	default:
 		break;
 	}
 }
 
+static void
+set_current_request (SoupConnection *conn, SoupMessage *req)
+{
+	g_return_if_fail (conn->priv->cur_req == NULL);
+
+	req->status = SOUP_MESSAGE_STATUS_RUNNING;
+	conn->priv->cur_req = req;
+	conn->priv->in_use = TRUE;
+	g_object_add_weak_pointer (G_OBJECT (req),
+				   (gpointer *)conn->priv->cur_req);
+}
+
+static void
+clear_current_request (SoupConnection *conn)
+{
+	if (conn->priv->cur_req) {
+		g_object_remove_weak_pointer (G_OBJECT (conn->priv->cur_req),
+					      (gpointer *)conn->priv->cur_req);
+		conn->priv->cur_req = NULL;
+	}
+	conn->priv->in_use = FALSE;
+}
+
 static void
 socket_disconnected (SoupSocket *sock, gpointer conn)
 {
@@ -309,6 +352,8 @@ tunnel_connect_finished (SoupMessage *msg, gpointer user_data)
 	SoupConnection *conn = user_data;
 	guint status = msg->status_code;
 
+	clear_current_request (conn);
+
 	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
 		if (!soup_socket_start_ssl (conn->priv->socket))
 			status = SOUP_STATUS_SSL_FAILED;
@@ -319,6 +364,37 @@ tunnel_connect_finished (SoupMessage *msg, gpointer user_data)
 	g_object_unref (msg);
 }
 
+static void
+tunnel_connect_restarted (SoupMessage *msg, gpointer user_data)
+{
+	SoupConnection *conn = user_data;
+	guint status = msg->status_code;
+
+	/* We only allow one restart: if another one happens, treat
+	 * it as "finished".
+	 */
+	g_signal_handlers_disconnect_by_func (msg, tunnel_connect_restarted, conn);
+	g_signal_connect (msg, "restarted",
+			  G_CALLBACK (tunnel_connect_finished), conn);
+
+	if (status == SOUP_STATUS_PROXY_AUTHENTICATION_REQUIRED) {
+		/* Our parent session has handled the authentication
+		 * and attempted to restart the message.
+		 */
+		if (soup_message_is_keepalive (msg)) {
+			/* Connection is still open, so just send the
+			 * message again.
+			 */
+			soup_connection_send_request (conn, msg);
+		} else {
+			/* Tell the session to try again. */
+			soup_message_set_status (msg, SOUP_STATUS_TRY_AGAIN);
+			soup_message_finished (msg);
+		}
+	} else
+		soup_message_finished (msg);
+}
+
 static void
 socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 {
@@ -335,11 +411,16 @@ socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 	}
 
 	/* See if we need to tunnel */
-	if (conn->priv->proxy_uri && conn->priv->origin_uri) {
+	if (conn->priv->proxy_uri &&
+	    conn->priv->origin_uri &&
+	    conn->priv->origin_uri->protocol == SOUP_PROTOCOL_HTTPS) {
 		SoupMessage *connect_msg;
 
 		connect_msg = soup_message_new_from_uri (SOUP_METHOD_CONNECT,
 							 conn->priv->origin_uri);
+
+		g_signal_connect (connect_msg, "restarted",
+				  G_CALLBACK (tunnel_connect_restarted), conn);
 		g_signal_connect (connect_msg, "finished",
 				  G_CALLBACK (tunnel_connect_finished), conn);
 
@@ -407,6 +488,9 @@ soup_connection_connect_sync (SoupConnection *conn)
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 		goto fail;
 
+	g_signal_connect (conn->priv->socket, "disconnected",
+			  G_CALLBACK (socket_disconnected), conn);
+
 	if (conn->priv->conn_uri->protocol == SOUP_PROTOCOL_HTTPS) {
 		if (!soup_socket_start_ssl (conn->priv->socket)) {
 			status = SOUP_STATUS_SSL_FAILED;
@@ -414,22 +498,40 @@ soup_connection_connect_sync (SoupConnection *conn)
 		}
 	}
 
-	if (conn->priv->proxy_uri && conn->priv->origin_uri) {
+	/* See if we need to tunnel */
+	if (conn->priv->proxy_uri &&
+	    conn->priv->origin_uri &&
+	    conn->priv->origin_uri->protocol == SOUP_PROTOCOL_HTTPS) {
 		SoupMessage *connect_msg;
 
 		connect_msg = soup_message_new_from_uri (SOUP_METHOD_CONNECT,
 							 conn->priv->origin_uri);
 		soup_connection_send_request (conn, connect_msg);
 		status = connect_msg->status_code;
+
+		if (status == SOUP_STATUS_PROXY_AUTHENTICATION_REQUIRED &&
+		    SOUP_MESSAGE_IS_STARTING (connect_msg)) {
+			if (soup_message_is_keepalive (connect_msg)) {
+				/* Try once more */
+				soup_connection_send_request (conn, connect_msg);
+				status = connect_msg->status_code;
+			} else
+				status = SOUP_STATUS_TRY_AGAIN;
+		}
+
 		g_object_unref (connect_msg);
 	}
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
 	fail:
-		g_object_unref (conn->priv->socket);
-		conn->priv->socket = NULL;
+		if (conn->priv->socket) {
+			g_object_unref (conn->priv->socket);
+			conn->priv->socket = NULL;
+		}
 	}
 
+	g_signal_emit (conn, signals[CONNECT_RESULT], 0,
+		       proxified_status (conn, status));
 	return proxified_status (conn, status);
 }
 
@@ -451,9 +553,51 @@ soup_connection_disconnect (SoupConnection *conn)
 
 	g_signal_handlers_disconnect_by_func (conn->priv->socket,
 					      socket_disconnected, conn);
+	soup_socket_disconnect (conn->priv->socket);
 	g_object_unref (conn->priv->socket);
 	conn->priv->socket = NULL;
 	g_signal_emit (conn, signals[DISCONNECTED], 0);
+
+	if (!conn->priv->cur_req ||
+	    conn->priv->cur_req->status_code != SOUP_STATUS_IO_ERROR)
+		return;
+
+	/* There was a message queued on this connection, but the
+	 * socket was closed while it was being sent.
+	 */
+
+	if (conn->priv->last_used != 0) {
+		/* If last_used is not 0, then that means at least one
+		 * message was successfully sent on this connection
+		 * before, and so the most likely cause of the
+		 * IO_ERROR is that the connection was idle for too
+		 * long and the server timed out and closed it (and we
+		 * didn't notice until after we started sending the
+		 * message). So we want the message to get tried again
+		 * on a new connection. The only code path that could
+		 * have gotten us to this point is through the call to
+		 * io_cleanup() in soup_message_io_finished(), and so
+		 * all we need to do to get the message requeued in
+		 * this case is to change its status.
+		 */
+		conn->priv->cur_req->status = SOUP_MESSAGE_STATUS_QUEUED;
+		return;
+	}
+
+	/* If conn->priv->last_used is 0, then that means this was the
+	 * first message to be sent on this connection, so the error
+	 * probably means that there's some network or server problem,
+	 * so we let the IO_ERROR be returned to the caller.
+	 *
+	 * Of course, it's also possible that the error in the
+	 * last_used != 0 case was because of a network/server problem
+	 * too. It's even possible that the message crashed the
+	 * server. In this case, requeuing it was the wrong thing to
+	 * do, but presumably, the next attempt will also get an
+	 * error, and eventually the message will be requeued onto a
+	 * fresh connection and get an error, at which point the error
+	 * will finally be returned to the caller.
+	 */
 }
 
 /**
@@ -467,7 +611,7 @@ soup_connection_is_in_use (SoupConnection *conn)
 {
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
 
-	return conn->priv->cur_req != NULL;
+	return conn->priv->in_use;
 }
 
 /**
@@ -486,35 +630,45 @@ soup_connection_last_used (SoupConnection *conn)
 }
 
 static void
-request_done (SoupMessage *req, gpointer user_data)
+request_restarted (SoupMessage *req, gpointer conn)
 {
-	SoupConnection *conn = user_data;
-
-	g_object_remove_weak_pointer (G_OBJECT (conn->priv->cur_req),
-				      (gpointer *)conn->priv->cur_req);
-	conn->priv->cur_req = NULL;
-	conn->priv->last_used = time (NULL);
-
-	g_signal_handlers_disconnect_by_func (req, request_done, conn);
-
 	if (!soup_message_is_keepalive (req))
 		soup_connection_disconnect (conn);
 }
 
+static void
+request_done (SoupMessage *req, gpointer user_data)
+{
+	SoupConnection *conn = user_data;
+
+	clear_current_request (conn);
+	conn->priv->last_used = time (NULL);
+
+	if (!soup_message_is_keepalive (req))
+		soup_connection_disconnect (conn);
+
+	g_signal_handlers_disconnect_by_func (req, request_done, conn);
+	g_signal_handlers_disconnect_by_func (req, request_restarted, conn);
+	g_object_unref (conn);
+}
+
 static void
 send_request (SoupConnection *conn, SoupMessage *req)
 {
-	if (req != conn->priv->cur_req) {
-		g_return_if_fail (conn->priv->cur_req == NULL);
-		conn->priv->cur_req = req;
-		g_object_add_weak_pointer (G_OBJECT (req),
-					   (gpointer *)conn->priv->cur_req);
+	g_object_ref (conn);
 
+	if (req != conn->priv->cur_req) {
+		set_current_request (conn, req);
+
+		g_signal_connect (req, "restarted",
+				  G_CALLBACK (request_restarted), conn);
 		g_signal_connect (req, "finished",
 				  G_CALLBACK (request_done), conn);
+
+		if (conn->priv->filter)
+			soup_message_filter_setup_message (conn->priv->filter, req);
 	}
 
-	soup_message_io_cancel (req);
 	soup_message_send_request (req, conn->priv->socket,
 				   conn->priv->proxy_uri != NULL);
 }
@@ -529,6 +683,14 @@ soup_connection_send_request (SoupConnection *conn, SoupMessage *req)
 	SOUP_CONNECTION_GET_CLASS (conn)->send_request (conn, req);
 }
 
+void
+soup_connection_reserve (SoupConnection *conn)
+{
+	g_return_if_fail (SOUP_IS_CONNECTION (conn));
+
+	conn->priv->in_use = TRUE;
+}
+
 void
 soup_connection_authenticate (SoupConnection *conn, SoupMessage *msg,
 			      const char *auth_type, const char *auth_realm,
